@@ -1,13 +1,12 @@
-#ifndef _OPENCV_FLANN_PRECOMP_HPP_
-#define _OPENCV_FLANN_PRECOMP_HPP_
+#ifndef OPENCV_FLANN_PRECOMP_HPP
+#define OPENCV_FLANN_PRECOMP_HPP
 
 #include <cstdio>
 #include <cstdarg>
 #include <sstream>
 
-# include "cvconfig.h"
-#include "opencv2/core/core.hpp"
-#include "opencv2/core/internal.hpp"
+#include "opencv2/core.hpp"
+#include "opencv2/core/utility.hpp"
 
 #include "opencv2/flann/miniflann.hpp"
 #include "opencv2/flann/dist.h"
@@ -21,4 +20,6 @@
 #include "opencv2/flann/all_indices.h"
 #include "opencv2/flann/flann_base.hpp"
 
+#include "opencv2/core/private.hpp"
+
 #endif
