@@ -1,24 +1,40 @@
 #!/usr/bin/env python
 
-import hdr_parser, sys, re, os, cStringIO
+from __future__ import print_function
+import hdr_parser, sys, re, os
 from string import Template
 
-gen_template_check_self = Template("""    if(!PyObject_TypeCheck(self, &pyopencv_${name}_Type))
+if sys.version_info[0] >= 3:
+    from io import StringIO
+else:
+    from cStringIO import StringIO
+
+ignored_arg_types = ["RNG*"]
+
+gen_template_check_self = Template("""    $cname* _self_ = NULL;
+    if(PyObject_TypeCheck(self, &pyopencv_${name}_Type))
+        _self_ = ${amp}((pyopencv_${name}_t*)self)->v${get};
+    if (_self_ == NULL)
         return failmsgp("Incorrect type of self (must be '${name}' or its derivative)");
-    $cname* _self_ = ${amp}((pyopencv_${name}_t*)self)->v;
 """)
 
-gen_template_check_self_algo = Template("""    if(!PyObject_TypeCheck(self, &pyopencv_${name}_Type))
+gen_template_check_self_algo = Template("""    $cname* _self_ = NULL;
+    if(PyObject_TypeCheck(self, &pyopencv_${name}_Type))
+        _self_ = dynamic_cast<$cname*>(${amp}((pyopencv_${name}_t*)self)->v.get());
+    if (_self_ == NULL)
         return failmsgp("Incorrect type of self (must be '${name}' or its derivative)");
-    $cname* _self_ = dynamic_cast<$cname*>(${amp}((pyopencv_${name}_t*)self)->v.obj);
 """)
 
-gen_template_call_constructor = Template("""self = PyObject_NEW(pyopencv_${name}_t, &pyopencv_${name}_Type);
+gen_template_call_constructor_prelude = Template("""self = PyObject_NEW(pyopencv_${name}_t, &pyopencv_${name}_Type);
         new (&(self->v)) Ptr<$cname>(); // init Ptr with placement new
-        if(self) ERRWRAP2(self->v = new $cname""")
+        if(self) """)
 
-gen_template_simple_call_constructor = Template("""self = PyObject_NEW(pyopencv_${name}_t, &pyopencv_${name}_Type);
-        if(self) ERRWRAP2(self->v = $cname""")
+gen_template_call_constructor = Template("""self->v.reset(new ${cname}${args})""")
+
+gen_template_simple_call_constructor_prelude = Template("""self = PyObject_NEW(pyopencv_${name}_t, &pyopencv_${name}_Type);
+        if(self) """)
+
+gen_template_simple_call_constructor = Template("""new (&(self->v)) ${cname}${args}""")
 
 gen_template_parse_args = Template("""const char* keywords[] = { $kw_list, NULL };
     if( PyArg_ParseTupleAndKeywords(args, kw, "$fmtspec", (char**)keywords, $parse_arglist)$code_cvt )""")
@@ -26,11 +42,26 @@ gen_template_parse_args = Template("""const char* keywords[] = { $kw_list, NULL
 gen_template_func_body = Template("""$code_decl
     $code_parse
     {
-        $code_fcall;
+        ${code_prelude}ERRWRAP2($code_fcall);
         $code_ret;
     }
 """)
 
+py_major_version = sys.version_info[0]
+if __name__ == "__main__":
+    if len(sys.argv) > 3:
+        if sys.argv[3] == 'PYTHON3':
+            py_major_version = 3
+        elif sys.argv[3] == 'PYTHON2':
+            py_major_version = 2
+        else:
+            raise Exception('Incorrect argument: expected PYTHON2 or PYTHON3, received: ' + sys.argv[3])
+if py_major_version >= 3:
+    head_init_str = "PyVarObject_HEAD_INIT(&PyType_Type, 0)"
+else:
+    head_init_str = """PyObject_HEAD_INIT(&PyType_Type)
+0,"""
+
 gen_template_simple_type_decl = Template("""
 struct pyopencv_${name}_t
 {
@@ -40,37 +71,37 @@ struct pyopencv_${name}_t
 
 static PyTypeObject pyopencv_${name}_Type =
 {
-    PyObject_HEAD_INIT(&PyType_Type)
-    0,
+    %s
     MODULESTR".$wname",
     sizeof(pyopencv_${name}_t),
 };
 
 static void pyopencv_${name}_dealloc(PyObject* self)
 {
+    ((pyopencv_${name}_t*)self)->v.${cname}::~${sname}();
     PyObject_Del(self);
 }
 
-static PyObject* pyopencv_from(const ${cname}& r)
+template<> PyObject* pyopencv_from(const ${cname}& r)
 {
     pyopencv_${name}_t *m = PyObject_NEW(pyopencv_${name}_t, &pyopencv_${name}_Type);
-    m->v = r;
+    new (&m->v) ${cname}(r); //Copy constructor
     return (PyObject*)m;
 }
 
-static bool pyopencv_to(PyObject* src, ${cname}& dst, const char* name="<unknown>")
+template<> bool pyopencv_to(PyObject* src, ${cname}& dst, const char* name)
 {
     if( src == NULL || src == Py_None )
         return true;
     if(!PyObject_TypeCheck(src, &pyopencv_${name}_Type))
     {
-        failmsg("Expected ${cname} for argument '%s'", name);
+        failmsg("Expected ${cname} for argument '%%s'", name);
         return false;
     }
     dst = ((pyopencv_${name}_t*)src)->v;
     return true;
 }
-""")
+""" % head_init_str)
 
 
 gen_template_type_decl = Template("""
@@ -82,8 +113,7 @@ struct pyopencv_${name}_t
 
 static PyTypeObject pyopencv_${name}_Type =
 {
-    PyObject_HEAD_INIT(&PyType_Type)
-    0,
+    %s
     MODULESTR".$wname",
     sizeof(pyopencv_${name}_t),
 };
@@ -94,7 +124,7 @@ static void pyopencv_${name}_dealloc(PyObject* self)
     PyObject_Del(self);
 }
 
-static PyObject* pyopencv_from(const Ptr<${cname}>& r)
+template<> PyObject* pyopencv_from(const Ptr<${cname}>& r)
 {
     pyopencv_${name}_t *m = PyObject_NEW(pyopencv_${name}_t, &pyopencv_${name}_Type);
     new (&(m->v)) Ptr<$cname1>(); // init Ptr with placement new
@@ -102,23 +132,23 @@ static PyObject* pyopencv_from(const Ptr<${cname}>& r)
     return (PyObject*)m;
 }
 
-static bool pyopencv_to(PyObject* src, Ptr<${cname}>& dst, const char* name="<unknown>")
+template<> bool pyopencv_to(PyObject* src, Ptr<${cname}>& dst, const char* name)
 {
     if( src == NULL || src == Py_None )
         return true;
     if(!PyObject_TypeCheck(src, &pyopencv_${name}_Type))
     {
-        failmsg("Expected ${cname} for argument '%s'", name);
+        failmsg("Expected ${cname} for argument '%%s'", name);
         return false;
     }
-    dst = ((pyopencv_${name}_t*)src)->v;
+    dst = ((pyopencv_${name}_t*)src)->v.dynamicCast<${cname}>();
     return true;
 }
 
-""")
+""" % head_init_str)
 
 gen_template_map_type_cvt = Template("""
-static bool pyopencv_to(PyObject* src, ${cname}& dst, const char* name="<unknown>");
+template<> bool pyopencv_to(PyObject* src, ${cname}& dst, const char* name);
 """)
 
 gen_template_set_prop_from_map = Template("""
@@ -174,7 +204,10 @@ static PyObject* pyopencv_${name}_get_${member}(pyopencv_${name}_t* p, void *clo
 gen_template_get_prop_algo = Template("""
 static PyObject* pyopencv_${name}_get_${member}(pyopencv_${name}_t* p, void *closure)
 {
-    return pyopencv_from(dynamic_cast<$cname*>(p->v.obj)${access}${member});
+    $cname* _self_ = dynamic_cast<$cname*>(p->v.get());
+    if (_self_ == NULL)
+        return failmsgp("Incorrect type of object (must be '${name}' or its derivative)");
+    return pyopencv_from(_self_${access}${member});
 }
 """)
 
@@ -198,7 +231,13 @@ static int pyopencv_${name}_set_${member}(pyopencv_${name}_t* p, PyObject *value
         PyErr_SetString(PyExc_TypeError, "Cannot delete the ${member} attribute");
         return -1;
     }
-    return pyopencv_to(value, dynamic_cast<$cname*>(p->v.obj)${access}${member}) ? 0 : -1;
+    $cname* _self_ = dynamic_cast<$cname*>(p->v.get());
+    if (_self_ == NULL)
+    {
+        failmsgp("Incorrect type of object (must be '${name}' or its derivative)");
+        return -1;
+    }
+    return pyopencv_to(value, _self_${access}${member}) ? 0 : -1;
 }
 """)
 
@@ -211,6 +250,7 @@ gen_template_rw_prop_init = Template("""
 
 simple_argtype_mapping = {
     "bool": ("bool", "b", "0"),
+    "size_t": ("size_t", "I", "0"),
     "int": ("int", "i", "0"),
     "float": ("float", "f", "0.f"),
     "double": ("double", "d", "0"),
@@ -232,26 +272,31 @@ class ClassInfo(object):
     def __init__(self, name, decl=None):
         self.cname = name.replace(".", "::")
         self.name = self.wname = normalize_class_name(name)
+        self.sname = name[name.rfind('.') + 1:]
         self.ismap = False
         self.issimple = False
         self.isalgorithm = False
         self.methods = {}
         self.props = []
         self.consts = {}
+        self.base = None
         customname = False
 
         if decl:
-            self.bases = decl[1].split()[1:]
-            if len(self.bases) > 1:
-                print "Note: Class %s has more than 1 base class (not supported by Python C extensions)" % (self.name,)
-                print "      Bases: ", " ".join(self.bases)
-                print "      Only the first base class will be used"
-                self.bases = [self.bases[0].strip(",")]
+            bases = decl[1].split()[1:]
+            if len(bases) > 1:
+                print("Note: Class %s has more than 1 base class (not supported by Python C extensions)" % (self.name,))
+                print("      Bases: ", " ".join(bases))
+                print("      Only the first base class will be used")
                 #return sys.exit(-1)
-            if self.bases and self.bases[0].startswith("cv::"):
-                self.bases[0] = self.bases[0][4:]
-            if self.bases and self.bases[0] == "Algorithm":
-                self.isalgorithm = True
+            elif len(bases) == 1:
+                self.base = bases[0].strip(",")
+                if self.base.startswith("cv::"):
+                    self.base = self.base[4:]
+                if self.base == "Algorithm":
+                    self.isalgorithm = True
+                self.base = self.base.replace("::", "_")
+
             for m in decl[2]:
                 if m.startswith("="):
                     self.wname = m[1:]
@@ -268,8 +313,8 @@ class ClassInfo(object):
     def gen_map_code(self, all_classes):
         code = "static bool pyopencv_to(PyObject* src, %s& dst, const char* name)\n{\n    PyObject* tmp;\n    bool ok;\n" % (self.cname)
         code += "".join([gen_template_set_prop_from_map.substitute(propname=p.name,proptype=p.tp) for p in self.props])
-        if self.bases:
-            code += "\n    return pyopencv_to(src, (%s&)dst, name);\n}\n" % all_classes[self.bases[0]].cname
+        if self.base:
+            code += "\n    return pyopencv_to(src, (%s&)dst, name);\n}\n" % all_classes[self.base].cname
         else:
             code += "\n    return true;\n}\n"
         return code
@@ -278,8 +323,8 @@ class ClassInfo(object):
         if self.ismap:
             return self.gen_map_code(all_classes)
 
-        getset_code = cStringIO.StringIO()
-        getset_inits = cStringIO.StringIO()
+        getset_code = StringIO()
+        getset_inits = StringIO()
 
         sorted_props = [(p.name, p) for p in self.props]
         sorted_props.sort()
@@ -302,10 +347,10 @@ class ClassInfo(object):
                     getset_code.write(gen_template_set_prop.substitute(name=self.name, member=pname, membertype=p.tp, access=access_op))
                 getset_inits.write(gen_template_rw_prop_init.substitute(name=self.name, member=pname))
 
-        methods_code = cStringIO.StringIO()
-        methods_inits = cStringIO.StringIO()
+        methods_code = StringIO()
+        methods_inits = StringIO()
 
-        sorted_methods = self.methods.items()
+        sorted_methods = list(self.methods.items())
         sorted_methods.sort()
 
         for mname, m in sorted_methods:
@@ -313,8 +358,8 @@ class ClassInfo(object):
             methods_inits.write(m.get_tab_entry())
 
         baseptr = "NULL"
-        if self.bases and all_classes.has_key(self.bases[0]):
-            baseptr = "&pyopencv_" + all_classes[self.bases[0]].name + "_Type"
+        if self.base and self.base in all_classes:
+            baseptr = "&pyopencv_" + all_classes[self.base].name + "_Type"
 
         code = gen_template_type_impl.substitute(name=self.name, wname=self.wname, cname=self.cname,
             getset_code=getset_code.getvalue(), getset_inits=getset_inits.getvalue(),
@@ -324,19 +369,15 @@ class ClassInfo(object):
         return code
 
 
-class ConstInfo(object):
-    def __init__(self, name, val):
-        self.cname = name.replace(".", "::")
-        self.name = re.sub(r"^cv\.", "", name).replace(".", "_")
-        if self.name.startswith("Cv"):
-            self.name = self.name[2:]
-        self.name = re.sub(r"([a-z])([A-Z])", r"\1_\2", self.name)
-        self.name = self.name.upper()
-        self.value = val
+def handle_ptr(tp):
+    if tp.startswith('Ptr_'):
+        tp = 'Ptr<' + "::".join(tp.split('_')[1:]) + '>'
+    return tp
+
 
 class ArgInfo(object):
     def __init__(self, arg_tuple):
-        self.tp = arg_tuple[0]
+        self.tp = handle_ptr(arg_tuple[0])
         self.name = arg_tuple[1]
         self.defval = arg_tuple[2]
         self.isarray = False
@@ -353,6 +394,7 @@ class ArgInfo(object):
             elif m == "/IO":
                 self.inputarg = True
                 self.outputarg = True
+                self.returnarg = True
             elif m.startswith("/A"):
                 self.isarray = True
                 self.arraylen = m[2:].strip()
@@ -363,7 +405,8 @@ class ArgInfo(object):
         self.py_outputarg = False
 
     def isbig(self):
-        return self.tp == "Mat" or self.tp == "vector_Mat"# or self.tp.startswith("vector")
+        return self.tp == "Mat" or self.tp == "vector_Mat"\
+               or self.tp == "UMat" or self.tp == "vector_UMat" # or self.tp.startswith("vector")
 
     def crepr(self):
         return "ArgInfo(\"%s\", %d)" % (self.name, self.outputarg)
@@ -374,13 +417,10 @@ class FuncVariant(object):
         self.classname = classname
         self.name = self.wname = name
         self.isconstructor = isconstructor
-        if self.isconstructor:
-            if self.wname.startswith("Cv"):
-                self.wname = self.wname[2:]
-            else:
-                self.wname = self.classname
 
-        self.rettype = decl[1]
+        self.docstring = decl[5]
+
+        self.rettype = decl[4] or handle_ptr(decl[1])
         if self.rettype == "void":
             self.rettype = ""
         self.args = []
@@ -424,9 +464,11 @@ class FuncVariant(object):
             argno += 1
             if a.name in self.array_counters:
                 continue
+            if a.tp in ignored_arg_types:
+                continue
             if a.returnarg:
                 outlist.append((a.name, argno))
-            if (not a.inputarg or a.returnarg) and a.isbig():
+            if (not a.inputarg) and a.isbig():
                 outarr_list.append((a.name, argno))
                 continue
             if not a.inputarg:
@@ -467,7 +509,7 @@ class FuncVariant(object):
         else:
             outstr = "None"
 
-        self.py_docstring = "%s(%s) -> %s" % (self.wname, argstr, outstr)
+        self.py_prototype = "%s(%s) -> %s" % (self.wname, argstr, outstr)
         self.py_noptargs = noptargs
         self.py_arglist = arglist
         for aname, argno in arglist:
@@ -479,11 +521,12 @@ class FuncVariant(object):
 
 
 class FuncInfo(object):
-    def __init__(self, classname, name, cname, isconstructor):
+    def __init__(self, classname, name, cname, isconstructor, namespace):
         self.classname = classname
         self.name = name
         self.cname = cname
         self.isconstructor = isconstructor
+        self.namespace = namespace
         self.variants = []
 
     def add_variant(self, decl):
@@ -497,7 +540,7 @@ class FuncInfo(object):
                 name = "getelem"
         else:
             classname = ""
-        return "pyopencv_" + classname + name
+        return "pyopencv_" + self.namespace.replace('.','_') + '_' + classname + name
 
     def get_wrapper_prototype(self):
         full_fname = self.get_wrapper_name()
@@ -508,32 +551,54 @@ class FuncInfo(object):
         return "static PyObject* %s(PyObject* %s, PyObject* args, PyObject* kw)" % (full_fname, self_arg)
 
     def get_tab_entry(self):
+        prototype_list = []
         docstring_list = []
+
         have_empty_constructor = False
         for v in self.variants:
-            s = v.py_docstring
+            s = v.py_prototype
             if (not v.py_arglist) and self.isconstructor:
                 have_empty_constructor = True
-            if s not in docstring_list:
-                docstring_list.append(s)
+            if s not in prototype_list:
+                prototype_list.append(s)
+                docstring_list.append(v.docstring)
+
         # if there are just 2 constructors: default one and some other,
         # we simplify the notation.
         # Instead of ClassName(args ...) -> object or ClassName() -> object
         # we write ClassName([args ...]) -> object
         if have_empty_constructor and len(self.variants) == 2:
             idx = self.variants[1].py_arglist != []
-            s = self.variants[idx].py_docstring
+            s = self.variants[idx].py_prototype
             p1 = s.find("(")
             p2 = s.rfind(")")
-            docstring_list = [s[:p1+1] + "[" + s[p1+1:p2] + "]" + s[p2:]]
+            prototype_list = [s[:p1+1] + "[" + s[p1+1:p2] + "]" + s[p2:]]
 
-        return Template('    {"$py_funcname", (PyCFunction)$wrap_funcname, METH_KEYWORDS, "$py_docstring"},\n'
+        # The final docstring will be: Each prototype, followed by
+        # their relevant doxygen comment
+        full_docstring = ""
+        for prototype, body in zip(prototype_list, docstring_list):
+            full_docstring += Template("$prototype\n$docstring\n\n\n\n").substitute(
+                prototype=prototype,
+                docstring='\n'.join(
+                    ['.   ' + line
+                     for line in body.split('\n')]
+                )
+            )
+
+        # Escape backslashes, newlines, and double quotes
+        full_docstring = full_docstring.strip().replace("\\", "\\\\").replace('\n', '\\n').replace("\"", "\\\"")
+        # Convert unicode chars to xml representation, but keep as string instead of bytes
+        full_docstring = full_docstring.encode('ascii', errors='xmlcharrefreplace').decode()
+
+        return Template('    {"$py_funcname", (PyCFunction)$wrap_funcname, METH_VARARGS | METH_KEYWORDS, "$py_docstring"},\n'
                         ).substitute(py_funcname = self.variants[0].wname, wrap_funcname=self.get_wrapper_name(),
-                                     py_docstring = "  or  ".join(docstring_list))
+                                     py_docstring = full_docstring)
 
     def gen_code(self, all_classes):
         proto = self.get_wrapper_prototype()
         code = "%s\n{\n" % (proto,)
+        code += "    using namespace %s;\n\n" % self.namespace.replace('.', '::')
 
         selfinfo = ClassInfo("")
         ismethod = self.classname != "" and not self.isconstructor
@@ -543,39 +608,22 @@ class FuncInfo(object):
         if self.classname:
             selfinfo = all_classes[self.classname]
             if not self.isconstructor:
-                amp = ""
-                if selfinfo.issimple:
-                    amp = "&"
+                amp = "&" if selfinfo.issimple else ""
                 if selfinfo.isalgorithm:
                     code += gen_template_check_self_algo.substitute(name=selfinfo.name, cname=selfinfo.cname, amp=amp)
                 else:
-                    code += gen_template_check_self.substitute(name=selfinfo.name, cname=selfinfo.cname, amp=amp)
+                    get = "" if selfinfo.issimple else ".get()"
+                    code += gen_template_check_self.substitute(name=selfinfo.name, cname=selfinfo.cname, amp=amp, get=get)
                 fullname = selfinfo.wname + "." + fullname
 
         all_code_variants = []
         declno = -1
         for v in self.variants:
             code_decl = ""
-            code_fcall = ""
             code_ret = ""
             code_cvt_list = []
 
-            if self.isconstructor:
-                code_decl += "    pyopencv_%s_t* self = 0;\n" % selfinfo.name
-                templ = gen_template_call_constructor
-                if selfinfo.issimple:
-                    templ = gen_template_simple_call_constructor
-                code_fcall = templ.substitute(name=selfinfo.name, cname=selfinfo.cname)
-            else:
-                code_fcall = "ERRWRAP2( "
-                if v.rettype:
-                    code_decl += "    " + v.rettype + " retval;\n"
-                    code_fcall += "retval = "
-                if ismethod:
-                    code_fcall += "_self_->" + self.cname
-                else:
-                    code_fcall += self.cname
-            code_fcall += "("
+            code_args = "("
             all_cargs = []
             parse_arglist = []
 
@@ -584,6 +632,16 @@ class FuncInfo(object):
             # form the function/method call,
             # for the list of type mappings
             for a in v.args:
+                if a.tp in ignored_arg_types:
+                    defval = a.defval
+                    if not defval and a.tp.endswith("*"):
+                        defval = 0
+                    assert defval
+                    if not code_args.endswith("("):
+                        code_args += ", "
+                    code_args += defval
+                    all_cargs.append([[None, ""], ""])
+                    continue
                 tp1 = tp = a.tp
                 amp = ""
                 defval0 = ""
@@ -594,7 +652,7 @@ class FuncInfo(object):
                         defval0 = "0"
                         tp1 = tp.replace("*", "_ptr")
                 if tp1.endswith("*"):
-                    print "Error: type with star: a.tp=%s, tp=%s, tp1=%s" % (a.tp, tp, tp1)
+                    print("Error: type with star: a.tp=%s, tp=%s, tp1=%s" % (a.tp, tp, tp1))
                     sys.exit(-1)
 
                 amapping = simple_argtype_mapping.get(tp, (tp, "O", defval0))
@@ -603,13 +661,20 @@ class FuncInfo(object):
                     if amapping[1] == "O":
                         code_decl += "    PyObject* pyobj_%s = NULL;\n" % (a.name,)
                         parse_name = "pyobj_" + a.name
-                        code_cvt_list.append("pyopencv_to(pyobj_%s, %s, %s)" % (a.name, a.name, a.crepr()))
+                        if a.tp == 'char':
+                            code_cvt_list.append("convert_to_char(pyobj_%s, &%s, %s)"% (a.name, a.name, a.crepr()))
+                        else:
+                            code_cvt_list.append("pyopencv_to(pyobj_%s, %s, %s)" % (a.name, a.name, a.crepr()))
 
                 all_cargs.append([amapping, parse_name])
 
                 defval = a.defval
                 if not defval:
                     defval = amapping[2]
+                else:
+                    if "UMat" in tp:
+                        if "Mat" in defval and "UMat" not in defval:
+                            defval = defval.replace("Mat", "UMat")
                 # "tp arg = tp();" is equivalent to "tp arg;" in the case of complex types
                 if defval == tp + "()" and amapping[1] == "O":
                     defval = ""
@@ -620,11 +685,34 @@ class FuncInfo(object):
                 else:
                     code_decl += "    %s %s;\n" % (amapping[0], a.name)
 
-                if not code_fcall.endswith("("):
-                    code_fcall += ", "
-                code_fcall += amp + a.name
+                if not code_args.endswith("("):
+                    code_args += ", "
+                code_args += amp + a.name
 
-            code_fcall += "))"
+            code_args += ")"
+
+            if self.isconstructor:
+                code_decl += "    pyopencv_%s_t* self = 0;\n" % selfinfo.name
+                if selfinfo.issimple:
+                    templ_prelude = gen_template_simple_call_constructor_prelude
+                    templ = gen_template_simple_call_constructor
+                else:
+                    templ_prelude = gen_template_call_constructor_prelude
+                    templ = gen_template_call_constructor
+
+                code_prelude = templ_prelude.substitute(name=selfinfo.name, cname=selfinfo.cname)
+                code_fcall = templ.substitute(name=selfinfo.name, cname=selfinfo.cname, args=code_args)
+            else:
+                code_prelude = ""
+                code_fcall = ""
+                if v.rettype:
+                    code_decl += "    " + v.rettype + " retval;\n"
+                    code_fcall += "retval = "
+                if ismethod:
+                    code_fcall += "_self_->" + self.cname
+                else:
+                    code_fcall += self.cname
+                code_fcall += code_args
 
             if code_cvt_list:
                 code_cvt_list = [""] + code_cvt_list
@@ -677,7 +765,7 @@ class FuncInfo(object):
                     (fmtspec, ", ".join(["pyopencv_from(" + aname + ")" for aname, argno in v.py_outlist]))
 
             all_code_variants.append(gen_template_func_body.substitute(code_decl=code_decl,
-                code_parse=code_parse, code_fcall=code_fcall, code_ret=code_ret))
+                code_parse=code_parse, code_prelude=code_prelude, code_fcall=code_fcall, code_ret=code_ret))
 
         if len(all_code_variants)==1:
             # if the function/method has only 1 signature, then just put it
@@ -689,19 +777,25 @@ class FuncInfo(object):
         return code
 
 
+class Namespace(object):
+    def __init__(self):
+        self.funcs = {}
+        self.consts = {}
+
+
 class PythonWrapperGenerator(object):
     def __init__(self):
         self.clear()
 
     def clear(self):
         self.classes = {}
-        self.funcs = {}
+        self.namespaces = {}
         self.consts = {}
-        self.code_types = cStringIO.StringIO()
-        self.code_funcs = cStringIO.StringIO()
-        self.code_func_tab = cStringIO.StringIO()
-        self.code_type_reg = cStringIO.StringIO()
-        self.code_const_reg = cStringIO.StringIO()
+        self.code_include = StringIO()
+        self.code_types = StringIO()
+        self.code_funcs = StringIO()
+        self.code_type_reg = StringIO()
+        self.code_ns_reg = StringIO()
         self.class_idx = 0
 
     def add_class(self, stype, name, decl):
@@ -709,69 +803,92 @@ class PythonWrapperGenerator(object):
         classinfo.decl_idx = self.class_idx
         self.class_idx += 1
 
-        if self.classes.has_key(classinfo.name):
-            print "Generator error: class %s (cname=%s) already exists" \
-                % (classinfo.name, classinfo.cname)
+        if classinfo.name in self.classes:
+            print("Generator error: class %s (cname=%s) already exists" \
+                % (classinfo.name, classinfo.cname))
             sys.exit(-1)
         self.classes[classinfo.name] = classinfo
-        if classinfo.bases and not classinfo.isalgorithm:
-            classinfo.isalgorithm = self.classes[classinfo.bases[0]].isalgorithm
+
+    def split_decl_name(self, name):
+        chunks = name.split('.')
+        namespace = chunks[:-1]
+        classes = []
+        while namespace and '.'.join(namespace) not in self.parser.namespaces:
+            classes.insert(0, namespace.pop())
+        return namespace, classes, chunks[-1]
+
 
     def add_const(self, name, decl):
-        constinfo = ConstInfo(name, decl[1])
-
-        if self.consts.has_key(constinfo.name):
-            print "Generator error: constant %s (cname=%s) already exists" \
-                % (constinfo.name, constinfo.cname)
+        cname = name.replace('.','::')
+        namespace, classes, name = self.split_decl_name(name)
+        namespace = '.'.join(namespace)
+        name = '_'.join(classes+[name])
+        ns = self.namespaces.setdefault(namespace, Namespace())
+        if name in ns.consts:
+            print("Generator error: constant %s (cname=%s) already exists" \
+                % (name, cname))
             sys.exit(-1)
-        self.consts[constinfo.name] = constinfo
+        ns.consts[name] = cname
 
     def add_func(self, decl):
-        classname = bareclassname = ""
-        name = decl[0]
-        dpos = name.rfind(".")
-        if dpos >= 0 and name[:dpos] != "cv":
-            classname = bareclassname = re.sub(r"^cv\.", "", name[:dpos])
-            name = name[dpos+1:]
-            dpos = classname.rfind(".")
-            if dpos >= 0:
-                bareclassname = classname[dpos+1:]
-                classname = classname.replace(".", "_")
-        cname = name
-        name = re.sub(r"^cv\.", "", name)
-        isconstructor = cname == bareclassname
-        cname = cname.replace(".", "::")
+        namespace, classes, barename = self.split_decl_name(decl[0])
+        cname = "::".join(namespace+classes+[barename])
+        name = barename
+        classname = ''
+        bareclassname = ''
+        if classes:
+            classname = normalize_class_name('.'.join(namespace+classes))
+            bareclassname = classes[-1]
+        namespace = '.'.join(namespace)
+
+        isconstructor = name == bareclassname
         isclassmethod = False
-        customname = False
         for m in decl[2]:
             if m == "/S":
                 isclassmethod = True
             elif m.startswith("="):
                 name = m[1:]
-                customname = True
-        func_map = self.funcs
+        if isclassmethod:
+            name = "_".join(classes+[name])
+            classname = ''
+        elif isconstructor:
+            name = "_".join(classes[:-1]+[name])
 
-        if not classname or isconstructor:
-            pass
-        elif isclassmethod:
-            if not customname:
-                name = classname + "_" + name
-            cname = classname + "::" + cname
-            classname = ""
+        if classname and not isconstructor:
+            cname = barename
+            func_map = self.classes[classname].methods
         else:
-            classinfo = self.classes.get(classname, ClassInfo(""))
-            if not classinfo.name:
-                print "Generator error: the class for method %s is missing" % (name,)
-                sys.exit(-1)
-            func_map = classinfo.methods
+            func_map = self.namespaces.setdefault(namespace, Namespace()).funcs
 
-        func = func_map.get(name, FuncInfo(classname, name, cname, isconstructor))
+        func = func_map.setdefault(name, FuncInfo(classname, name, cname, isconstructor, namespace))
         func.add_variant(decl)
-        if len(func.variants) == 1:
-            func_map[name] = func
 
-    def gen_const_reg(self, constinfo):
-        self.code_const_reg.write("PUBLISH2(%s,%s);\n" % (constinfo.name, constinfo.cname))
+
+    def gen_namespace(self, ns_name):
+        ns = self.namespaces[ns_name]
+        wname = normalize_class_name(ns_name)
+
+        self.code_ns_reg.write('static PyMethodDef methods_%s[] = {\n'%wname)
+        for name, func in sorted(ns.funcs.items()):
+            self.code_ns_reg.write(func.get_tab_entry())
+        self.code_ns_reg.write('    {NULL, NULL}\n};\n\n')
+
+        self.code_ns_reg.write('static ConstDef consts_%s[] = {\n'%wname)
+        for name, cname in sorted(ns.consts.items()):
+            self.code_ns_reg.write('    {"%s", %s},\n'%(name, cname))
+            compat_name = re.sub(r"([a-z])([A-Z])", r"\1_\2", name).upper()
+            if name != compat_name:
+                self.code_ns_reg.write('    {"%s", %s},\n'%(compat_name, cname))
+        self.code_ns_reg.write('    {NULL, 0}\n};\n\n')
+
+    def gen_namespaces_reg(self):
+        self.code_ns_reg.write('static void init_submodules(PyObject * root) \n{\n')
+        for ns_name in sorted(self.namespaces):
+            if ns_name.split('.')[0] == 'cv':
+                wname = normalize_class_name(ns_name)
+                self.code_ns_reg.write('  init_submodule(root, MODULESTR"%s", methods_%s, consts_%s);\n' % (ns_name[2:], wname, wname))
+        self.code_ns_reg.write('};\n')
+
 
     def save(self, path, name, buf):
         f = open(path + "/" + name, "wt")
@@ -780,11 +897,14 @@ class PythonWrapperGenerator(object):
 
     def gen(self, srcfiles, output_path):
         self.clear()
-        parser = hdr_parser.CppHeaderParser()
+        self.parser = hdr_parser.CppHeaderParser(generate_umat_decls=True)
 
         # step 1: scan the headers and build more descriptive maps of classes, consts, functions
         for hdr in srcfiles:
-            decls = parser.parse(hdr)
+            decls = self.parser.parse(hdr)
+            if len(decls) == 0:
+                continue
+            self.code_include.write( '#include "{0}"\n'.format(hdr[hdr.rindex('opencv2/'):]) )
             for decl in decls:
                 name = decl[0]
                 if name.startswith("struct") or name.startswith("class"):
@@ -800,8 +920,24 @@ class PythonWrapperGenerator(object):
                     # function
                     self.add_func(decl)
 
+        # step 1.5 check if all base classes exist
+        for name, classinfo in self.classes.items():
+            if classinfo.base:
+                chunks = classinfo.base.split('_')
+                base = '_'.join(chunks)
+                while base not in self.classes and len(chunks)>1:
+                    del chunks[-2]
+                    base = '_'.join(chunks)
+                if base not in self.classes:
+                    print("Generator error: unable to resolve base %s for %s"
+                        % (classinfo.base, classinfo.name))
+                    sys.exit(-1)
+                classinfo.base = base
+                classinfo.isalgorithm |= self.classes[base].isalgorithm
+                self.classes[name] = classinfo
+
         # step 2: generate code for the classes and their methods
-        classlist = self.classes.items()
+        classlist = list(self.classes.items())
         classlist.sort()
         for name, classinfo in classlist:
             if classinfo.ismap:
@@ -811,7 +947,7 @@ class PythonWrapperGenerator(object):
                     templ = gen_template_simple_type_decl
                 else:
                     templ = gen_template_type_decl
-                self.code_types.write(templ.substitute(name=name, wname=classinfo.wname, cname=classinfo.cname,
+                self.code_types.write(templ.substitute(name=name, wname=classinfo.wname, cname=classinfo.cname, sname=classinfo.sname,
                                       cname1=("cv::Algorithm" if classinfo.isalgorithm else classinfo.cname)))
 
         # register classes in the same order as they have been declared.
@@ -826,25 +962,27 @@ class PythonWrapperGenerator(object):
                 self.code_type_reg.write("MKTYPE2(%s);\n" % (classinfo.name,) )
 
         # step 3: generate the code for all the global functions
-        funclist = self.funcs.items()
-        funclist.sort()
-        for name, func in funclist:
-            code = func.gen_code(self.classes)
-            self.code_funcs.write(code)
-            self.code_func_tab.write(func.get_tab_entry())
+        for ns_name, ns in sorted(self.namespaces.items()):
+            if ns_name.split('.')[0] != 'cv':
+                continue
+            for name, func in sorted(ns.funcs.items()):
+                code = func.gen_code(self.classes)
+                self.code_funcs.write(code)
+            self.gen_namespace(ns_name)
+        self.gen_namespaces_reg()
 
         # step 4: generate the code for constants
-        constlist = self.consts.items()
+        constlist = list(self.consts.items())
         constlist.sort()
         for name, constinfo in constlist:
             self.gen_const_reg(constinfo)
 
         # That's it. Now save all the files
+        self.save(output_path, "pyopencv_generated_include.h", self.code_include)
         self.save(output_path, "pyopencv_generated_funcs.h", self.code_funcs)
-        self.save(output_path, "pyopencv_generated_func_tab.h", self.code_func_tab)
-        self.save(output_path, "pyopencv_generated_const_reg.h", self.code_const_reg)
         self.save(output_path, "pyopencv_generated_types.h", self.code_types)
         self.save(output_path, "pyopencv_generated_type_reg.h", self.code_type_reg)
+        self.save(output_path, "pyopencv_generated_ns_reg.h", self.code_ns_reg)
 
 if __name__ == "__main__":
     srcfiles = hdr_parser.opencv_hdr_list
@@ -852,6 +990,6 @@ if __name__ == "__main__":
     if len(sys.argv) > 1:
         dstdir = sys.argv[1]
     if len(sys.argv) > 2:
-        srcfiles = sys.argv[2:]
+        srcfiles = [f.strip() for f in open(sys.argv[2], 'r').readlines()]
     generator = PythonWrapperGenerator()
     generator.gen(srcfiles, dstdir)
